// SPDX-License-Identifier: BSD-3-Clause

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <time.h>

#include "os_graph.h"
#include "os_threadpool.h"
#include "log/log.h"
#include "utils.h"

#define NUM_THREADS		4

static int sum;
static os_graph_t *graph;
static os_threadpool_t *tp;
/* Define graph synchronization mechanisms. */
static pthread_mutex_t graph_lock;
static pthread_mutex_t sum_lock;

void os_destroy_arg(void *arg)
{
	free(arg);
}

static void process_node(unsigned int idx);

/*  Define graph task argument. */
static void process_neighbours(void *arg)
{
	unsigned int idx = *(unsigned int *)arg;
	os_node_t *node = graph->nodes[idx];

	pthread_mutex_lock(&sum_lock);
	sum += node->info;
	pthread_mutex_unlock(&sum_lock);

	pthread_mutex_lock(&graph_lock);
	for (unsigned int i = 0; i < node->num_neighbours; ++i)
		if (graph->visited[node->neighbours[i]] == NOT_VISITED)
			process_node(node->neighbours[i]);
	pthread_mutex_unlock(&graph_lock);
}

static void process_node(unsigned int idx)
{
	/* Implement thread-pool based processing of graph. */
	unsigned int *new_arg = malloc(sizeof(unsigned int));

	*new_arg = idx;
	graph->visited[idx] = DONE;
	os_task_t *new_task = create_task(process_neighbours, (void *)new_arg, os_destroy_arg);

	enqueue_task(tp, new_task);
}

int main(int argc, char *argv[])
{
	FILE *input_file;

	if (argc != 2) {
		fprintf(stderr, "Usage: %s input_file\n", argv[0]);
		exit(EXIT_FAILURE);
	}

	input_file = fopen(argv[1], "r");
	DIE(input_file == NULL, "fopen");

	graph = create_graph_from_file(input_file);

	/* Initialize graph synchronization mechanisms. */
	pthread_mutex_init(&graph_lock, NULL);
	pthread_mutex_init(&sum_lock, NULL);
	tp = create_threadpool(NUM_THREADS);
	process_node(0);
	wait_for_completion(tp);
	destroy_threadpool(tp);

	pthread_mutex_destroy(&graph_lock);
	pthread_mutex_destroy(&sum_lock);
	printf("%d", sum);

	return 0;
}
